/***************** Part of full PrhoLog with proximity relations ***************

  Prholog-full-prox is a  package for strategic rule-based programming
  with individual, function, sequence, and context variables and
  membership constraints over them. It extends Prholog-full by adding
  a strategy for proximity relations.

  compile.pl, written by Temur Kutsia and Besik Dundua.
  Compiles Rholog files into Prolog files.
  Last edit: September 24, 2021.
 
  Tested for SWI-Prolog (8.2.1 and later).

  (c) RISC, Johannes Kepler University Linz, Austria

********************************************************************************/

             /**************************************
             *           Compilation               *
             **************************************/

%--------------------------------------------------------------------------
%   Compile Rholog Clauses to Prolog
%--------------------------------------------------------------------------

compile_rholog_clause(end_of_file) :-
    !.
compile_rholog_clause(Rholog_clause) :-
    compile_rholog_clause_to_prolog(Rholog_clause, Prolog_clause),
    !,
    portray_clause(Prolog_clause).


%--------------------------------------------------------------------------
%   Compiling Rholog Clauses into Prolog Clauses
%--------------------------------------------------------------------------

compile_rholog_clause_to_prolog((Strategy :: Lhs ==> Rhs :- Body), Compiled) :-
    !,
    compile_rholog_clause_to_prolog((Strategy :: Lhs ==> Rhs where [] :- Body), 
        Compiled).

compile_rholog_clause_to_prolog((Strategy :: Lhs ==> Rhs 
    where Constraint :- Body),
    ( Clause_head :-
         solve([[hdg_rholog_internal|Str_args_list] << 
             [hdg_rholog_internal|Str_args_query]], Subst0),
         instance(Lhs_list, Subst0, Lhs_list_instance),
         solve([Lhs_list_instance << Lhs_query | Constraint_list], Subst1),
         Prolog_body,
         instance(Rhs_list, Subst2, Rhs_list_instance),
         solve([Rhs_query << Rhs_list_instance | Constraint_query], Subst3),
         compose(Subst_acc, Subst3, Subst)
    )
) :-
    !,
    term_constraints_to_symbol_list_constraints(Constraint, Constraint_list),
    hedge_to_term_list(Lhs, Lhs_list),
    hedge_to_term_list(Rhs, Rhs_list),
    compile_rholog_clause_body(Body, Subst0, Subst1, Subst2,
                      Lambda, T_norm, Degree_acc, Degree, Prolog_body),
    get_strategy_name_and_arguments(Strategy, Str_name, Str_args_list),
    assert(to_abolish_rholog_internal(Str_name, 10)),
    Clause_head =.. [Str_name, Str_args_query, Lhs_query, Rhs_query, Constraint_query,
                     Lambda, T_norm, Degree_acc, Degree, Subst_acc, Subst].
compile_rholog_clause_to_prolog((Strategy :: Lhs ==> Rhs), Compiled) :-
    !,
    compile_rholog_clause_to_prolog((Strategy :: Lhs ==> Rhs where []), 
        Compiled).
compile_rholog_clause_to_prolog((Strategy :: Lhs ==> Rhs where Constraint),
    ( Clause_head :-
         solve([[hdg_rholog_internal|Str_args_list] << 
      [hdg_rholog_internal|Str_args_query]], Subst0),
         instance(Lhs_list, Subst0, Lhs_list_instance),
         solve([Lhs_list_instance << Lhs_query | Constraint_list], Subst1),
         append(Subst0, Subst1, Subst2),
         instance(Rhs_list, Subst2, Rhs_list_instance),
         solve([Rhs_query << Rhs_list_instance | Constraint_query], Subst3),
         compose(Subst_acc, Subst3, Subst)
    )
) :-
    !,
    term_constraints_to_symbol_list_constraints(Constraint, Constraint_list),
    hedge_to_term_list(Lhs, Lhs_list),
    hedge_to_term_list(Rhs, Rhs_list),
    get_strategy_name_and_arguments(Strategy, Str_name, Str_args_list),
    assert(to_abolish_rholog_internal(Str_name, 10)),
    Clause_head =.. [Str_name, Str_args_query, Lhs_query, Rhs_query, 
        Constraint_query, _, _, Degree, Degree, Subst_acc, Subst].
compile_rholog_clause_to_prolog((Strategy_shortcut := Strategy_full), 
    Compiled
) :-
    !,
    compile_rholog_clause_to_prolog(
        (Strategy_shortcut :: s_X ==> s_Y :- Strategy_full :: s_X ==> s_Y),
        Compiled
    ).
compile_rholog_clause_to_prolog((:- X), (:- X)) :-
    X,
    !.
compile_rholog_clause_to_prolog(Prolog_clause, Prolog_clause).


%------------------------------------------------------------------------------
%   Compiling a Rholog clause body into a Prolog clause body
%------------------------------------------------------------------------------


compile_rholog_clause_body(Strategy :: Lhs ==> Rhs, Subst_strat_in, Subst_lhs_in,
    Subst_out, Lambda, T_norm, Degree_acc, Degree, Prolog_body
) :-
    !,
    compile_rholog_clause_body(Strategy :: Lhs ==> Rhs where [], Subst_strat_in,
                  Subst_lhs_in, Subst_out, Lambda, T_norm,
                  Degree_acc, Degree, Prolog_body).
compile_rholog_clause_body(Strategy :: Lhs ==> Rhs where Constraint, Subst_strat_in,
                  Subst_lhs_in, Subst_out, Lambda, T_norm, Degree_acc, Degree,
    ( append(Subst_strat_in, Subst_lhs_in, Subst_in),
      instance([Str_name|Str_args_list], Subst_in,
          [Str_name_instance|Str_args_list_instance]),
      instance(Lhs_list, Subst_in, Lhs_list_instance),
      instance(Rhs_list, Subst_in, Rhs_list_instance),
      instance(Constraint_list, Subst_in, Constraint_list_instance),
      get_strategy_name_arguments_from_the_list_form(
          Str_name_instance, Str_name_new, Str_args_list_1),
      append(Str_args_list_1, Str_args_list_instance, Str_args_list_new),
      Subgoal =..[Str_name_new, Str_args_list_new, Lhs_list_instance, 
          Rhs_list_instance, Constraint_list_instance, Lambda,
          T_norm, Degree_acc, Degree, Subst_in, Subst_out],
      Subgoal
    )
) :-
    !,
    term_constraints_to_symbol_list_constraints(Constraint, Constraint_list),
    hedge_to_term_list(Lhs, Lhs_list),
    hedge_to_term_list(Rhs, Rhs_list),
    get_strategy_name_and_arguments(Strategy, Str_name, Str_args_list).



compile_rholog_clause_body(Strategy :: Lhs =\=> Rhs, Subst_strat_in, 
    Subst_lhs_in, Subst_out, Lambda, T_norm, Degree_acc, Degree, Prolog_body
) :-
    !,
    compile_rholog_clause_body(Strategy :: Lhs =\=> Rhs where [],
        Subst_strat_in, Subst_lhs_in, Subst_out,
        Lambda, T_norm, Degree_acc, Degree, Prolog_body).
compile_rholog_clause_body(Strategy :: Lhs =\=> Rhs where Constraint,
    Subst_strat_in, Subst_lhs_in, Subst_out,
    Lambda, T_norm, Degree_acc, Degree,
    ( append(Subst_strat_in, Subst_lhs_in, Subst_in),
      instance([Str_name|Str_args_list], Subst_in,
          [Str_name_instance|Str_args_list_instance]),
      instance(Lhs_list, Subst_in, Lhs_list_instance),
      instance(Rhs_list, Subst_in, Rhs_list_instance),
      instance(Constraint_list, Subst_in, Constraint_list_instance),
      get_strategy_name_arguments_from_the_list_form(
          Str_name_instance, Str_name_new, Str_args_list_1),
      append(Str_args_list_1, Str_args_list_instance, Str_args_list_new),
      Subgoal =..[Str_name_new, Str_args_list_new, Lhs_list_instance,
          Rhs_list_instance, Constraint_list_instance, Lambda,
          T_norm, Degree_acc, Degree, Subst_in, Subst_out],
      \+(Subgoal),
      Subst_out = Subst_in
    )
) :-
    !,
    term_constraints_to_symbol_list_constraints(Constraint, Constraint_list),
    hedge_to_term_list(Lhs, Lhs_list),
    hedge_to_term_list(Rhs, Rhs_list),
    get_strategy_name_and_arguments(Strategy, Str_name, Str_args_list).



compile_rholog_clause_body((Rholog_subgoal1, Rholog_subgoal2),
    Subst_strat_in, Subst_lhs_in, Subst_out, Lambda, T_norm, Degree_acc, Degree,
    (Prolog_subgoal1, Prolog_subgoal2)
) :-
   !,
   compile_rholog_clause_body(Rholog_subgoal1, Subst_strat_in, Subst_lhs_in, 
       Subst, Lambda, T_norm, Degree_acc, Degree_temp, Prolog_subgoal1),
   compile_rholog_clause_body(Rholog_subgoal2, [], Subst, Subst_out,
       Lambda, T_norm, Degree_temp, Degree, Prolog_subgoal2).
compile_rholog_clause_body((Rholog_subgoal1 -> Rholog_subgoal2),
    Subst_strat_in, Subst_lhs_in, Subst_out,
    Lambda, T_norm, Degree_acc, Degree,
    (Prolog_subgoal1 -> Prolog_subgoal2)
) :-
   !,
   compile_rholog_clause_body(Rholog_subgoal1, Subst_strat_in, Subst_lhs_in, 
       Subst, Lambda, T_norm, Degree_acc, Degree_temp, Prolog_subgoal1),
   compile_rholog_clause_body(Rholog_subgoal2, [], Subst, Subst_out,
       Lambda, T_norm, Degree_temp, Degree, Prolog_subgoal2).
compile_rholog_clause_body((Rholog_subgoal1 *-> Rholog_subgoal2),
    Subst_strat_in, Subst_lhs_in, Subst_out,
    Lambda, T_norm, Degree_acc, Degree,
    (Prolog_subgoal1 *-> Prolog_subgoal2)
) :-
   !,
   compile_rholog_clause_body(Rholog_subgoal1, Subst_strat_in, Subst_lhs_in, 
       Subst, Lambda, T_norm, Degree_acc, Degree_temp, Prolog_subgoal1),
   compile_rholog_clause_body(Rholog_subgoal2, [], Subst, Subst_out, 
       Lambda, T_norm, Degree_temp, Degree, Prolog_subgoal2).
compile_rholog_clause_body((Rholog_subgoal1; Rholog_subgoal2),
   Subst_strat_in, Subst_lhs_in, Subst_out,
   Lambda, T_norm, Degree_acc, Degree,
   (Prolog_subgoal1; Prolog_subgoal2)
) :-
   !,
   compile_rholog_clause_body(Rholog_subgoal1, Subst_strat_in, Subst_lhs_in, 
       Subst_out, Lambda, T_norm, Degree_acc, Degree, Prolog_subgoal1),
   compile_rholog_clause_body(Rholog_subgoal2, Subst_strat_in, Subst_lhs_in, 
       Subst_out, Lambda, T_norm, Degree_acc, Degree, Prolog_subgoal2).
compile_rholog_clause_body(!, Subst_strat_in, Subst_lhs_in, Subst_out,
       _, _, Degree_acc, Degree_acc,
   (append(Subst_strat_in, Subst_lhs_in, Subst_out),!)
) :-
    !.
compile_rholog_clause_body(nl, Subst_strat_in, Subst_lhs_in, Subst_out,
     _, _,  Degree_acc, Degree_acc,
    ( append(Subst_strat_in, Subst_lhs_in, Subst_out),
      nl
    )
) :-
    !.
compile_rholog_clause_body(write(X), Subst_strat_in, Subst_lhs_in, Subst_in,
     _, _, Degree_acc, Degree_acc,
    ( append(Subst_strat_in, Subst_lhs_in, Subst_in),
      instance(Prolog_style_subgoal_list, Subst_in, 
          Prolog_style_subgoal_list_instance),
      symbol_list_to_term(Prolog_style_subgoal_list_instance, Prolog_subgoal),
          Prolog_subgoal
    )
) :-
    !,
    term_to_list(write(X), Prolog_style_subgoal_list).
compile_rholog_clause_body(Prolog_style_subgoal, Subst_strat_in, 
    Subst_lhs_in, Subst_out, _, _,  Degree_acc, Degree_acc,
    ( append(Subst_strat_in, Subst_lhs_in, Subst_in),
      instance(Prolog_style_subgoal_list, Subst_in,
          Prolog_style_subgoal_list_instance),
      replace_rholog_ind_vars_by_prolog_vars(
          Prolog_style_subgoal_list_instance,
          Prolog_style_subgoal_list_instance_with_Prolog_vars),
      symbol_list_to_term(
          Prolog_style_subgoal_list_instance_with_Prolog_vars,
          Prolog_subgoal),
      Prolog_subgoal,
      term_to_list(Prolog_subgoal, Prolog_subgoal_list),
      solve([Prolog_style_subgoal_list_instance <<
          Prolog_subgoal_list], Subst_temp),
      compose(Subst_in, Subst_temp, Subst_out)
    )
) :-
    !,
    term_to_list(Prolog_style_subgoal, Prolog_style_subgoal_list).


%--------------------------------------------------------------------------
%   Compiling Rholog Query into Prolog.
%   ? is the top-level predicate that executes Rholog goals.
%--------------------------------------------------------------------------

?(Query, Result):- ?(Query, 1, goedel, _, Result).

?(Query, Lambda, Degree, Result):- ?(Query, Lambda, goedel, Degree, Result).


?(Query, Lambda, T_norm, Degree, Result) :-
    catch(
        ( check_rholog_query_syntax(Query),
          compile_rholog_query(Query, Lambda, T_norm, Degree, Result)
        ),
        E,
        print_message(user_error,E)).

compile_rholog_query(Strategy :: Lhs ==> Rhs, Lambda, T_norm, Degree, Result) :-
    !,
    compile_rholog_query(Strategy :: Lhs ==> Rhs where [], Lambda, T_norm, Degree, Result).


compile_rholog_query(Strategy :: Lhs ==> Rhs where Constraint, Lambda, T_norm, Degree, Result) :-
    !,
    compile_rholog_clause_body(Strategy :: Lhs ==> Rhs where Constraint,
        [], [], Subst_out, Lambda, T_norm, 1, Degree, Compiled),
    Compiled,
    symbol_list_substitutions_to_term_substitutions(Subst_out, Result).


compile_rholog_query(Strategy :: Lhs =\=> Rhs, Lambda, T_norm, Degree, Result) :-
    !,
    compile_rholog_query(Strategy :: Lhs =\=> Rhs where [], Lambda, T_norm, Degree, Result).
compile_rholog_query(Strategy :: Lhs =\=> Rhs where Constraint, Lambda, T_norm, Degree, Result) :-
    !,
    compile_rholog_clause_body(Strategy :: Lhs =\=> Rhs where Constraint,
        [], [], Subst_out, Lambda, T_norm, 1, Degree, Compiled),
    Compiled,
    symbol_list_substitutions_to_term_substitutions(Subst_out, Result).
compile_rholog_query((Query1,Query2), Lambda, T_norm, Degree, Result) :-
    !,
    compile_rholog_clause_body((Query1,Query2), [], [], Subst_out,
         Lambda, T_norm, 1, Degree, Compiled),
    Compiled,
    symbol_list_substitutions_to_term_substitutions(Subst_out, Result).
compile_rholog_query((Query1;Query2), Lambda, T_norm, Degree, Result) :-
    !,
    compile_rholog_clause_body((Query1;Query2), [], [], Subst_out,
          Lambda, T_norm, 1, Degree, Compiled),
    Compiled,
    symbol_list_substitutions_to_term_substitutions(Subst_out, Result).
compile_rholog_query(Prolog_subquery, Lambda, T_norm, Degree, Result) :-
    !,
    compile_rholog_clause_body(Prolog_subquery,[], [], Subst_out,
    Lambda, T_norm, 1, Degree, Compiled),
    Compiled,
    symbol_list_substitutions_to_term_substitutions(Subst_out, Result).


%-------------------------------------------------------------------------
%   hedge_to_term_list(+Hedge, ?Term_list)
%   Succeeds when Term_list is a list of list representation
%   of the terms in Hedge. The first element in Term_list is the
%   reserved constant hdg_rholog_internal.
%   Example:
%      ?- hedge_to_term_list((f(a,b),1,h(g(b))), X).
%      X = [hdg_rholog_internal, [f, a, b], 1, [h, [g, b]]]
%------------------------------------------------------------------------- 

hedge_to_term_list(Hedge, Term_list) :-
    hedge_to_term_list(Hedge, [hdg_rholog_internal], Term_list).
hedge_to_term_list(eps, Acc, Acc) :-
    !.
hedge_to_term_list((Term1,Term2), Acc, Term_list) :-
    !,
    hedge_to_term_list(Term2, Acc, Acc1),
    hedge_to_term_list(Term1, Acc1, Term_list).
hedge_to_term_list(Single_term, [hdg_rholog_internal|T],
    [hdg_rholog_internal,Single_term_list_representation|T]
) :-
    term_to_list(Single_term, Single_term_list_representation).


%-------------------------------------------------------------------------
%    Get strategy name and arguments. 
%    Arguments are returned as a term list.
%------------------------------------------------------------------------- 

get_strategy_name_and_arguments(Strategy, Strategy_name, []) :-
    atomic(Strategy),
    !,
    transform_var_or_constant(Strategy, Strategy_name).
get_strategy_name_and_arguments(Strategy, Strategy_name, 
    Strategy_arguments_list
) :-
    term_to_list(Strategy, [Strategy_name|Strategy_arguments_list]).
       

%--------------------------------------------------------------------------
%    Get strategy name and arguments from the list form.
%--------------------------------------------------------------------------
       
get_strategy_name_arguments_from_the_list_form(Strategy, Strategy, []) :-
    atomic(Strategy),
    !.
get_strategy_name_arguments_from_the_list_form(
    [Strategy_name|Strategy_arguments_list], Strategy_name, 
    Strategy_arguments_list).


%---------------------------------------------------------------------------
%    Replace Rholog individual Variables by Prolog variables
%---------------------------------------------------------------------------

replace_rholog_ind_vars_by_prolog_vars(X, Y) :-
    replace_rholog_ind_vars_by_prolog_vars(X, [], Y).
replace_rholog_ind_vars_by_prolog_vars([], _, []) :-
    !.
replace_rholog_ind_vars_by_prolog_vars([[H|T]|Others], Replacement,
    [Head_repl|Others_repl]
) :-
    !,
    replace_rholog_ind_vars_by_prolog_vars([H|T], Replacement, Head_repl),
    replace_rholog_ind_vars_by_prolog_vars(Others, Replacement, Others_repl).
replace_rholog_ind_vars_by_prolog_vars([ind_var(X)|Others], Replacement,
    [Prolog_var|Others_repl]
) :-
    !,
    ind_var_to_prolog_var(ind_var(X), Replacement, Replacement_new, Prolog_var),
    replace_rholog_ind_vars_by_prolog_vars(Others, Replacement_new, 
        Others_repl).
replace_rholog_ind_vars_by_prolog_vars([i_|Others], Replacement,
    [Prolog_var|Others_repl]
) :-
    !,
    ind_var_to_prolog_var(i_, Replacement, Replacement_new, Prolog_var),
    replace_rholog_ind_vars_by_prolog_vars(Others, Replacement_new, 
        Others_repl).
replace_rholog_ind_vars_by_prolog_vars([H|_], _, _) :-
    functor(H, _, 1),
    !,
    throw(error(syntax_error(variables_in_prolog_clauses),_)).
replace_rholog_ind_vars_by_prolog_vars([H|Others], Replacement, 
    [H|Others_new]
) :-
    !,
    replace_rholog_ind_vars_by_prolog_vars(Others, Replacement, Others_new).
replace_rholog_ind_vars_by_prolog_vars(Atom, _, Atom).

ind_var_to_prolog_var(i_, Replacement, Replacement, _) :-
    !.
ind_var_to_prolog_var(Ind_var, [Ind_var--->Prolog_var|T],
    [Ind_var--->Prolog_var|T], Prolog_var
) :-
    !.
ind_var_to_prolog_var(Ind_var, [H|Tail], [H|Tail_new], Prolog_var) :-
    !,
    ind_var_to_prolog_var(Ind_var, Tail, Tail_new, Prolog_var).
ind_var_to_prolog_var(Ind_var, [], [Ind_var--->Prolog_var], Prolog_var).
